import string
from typing import TYPE_CHECKING, Any
from ape.contracts import ContractInstance
from ape.types import AddressType
from ape.utils import ManagerAccessMixin
from ape.utils.misc import cached_property
from eth_pydantic_types import abi
from eth_utils.crypto import keccak
from ethpm_types.abi import MethodABI
from pydantic import BaseModel, field_validator

if TYPE_CHECKING:
    from .main import Purse


class AccessoryMethod(BaseModel):
    method: abi.bytes4
    accessory: abi.address

    @field_validator("method", mode="before")
    def convert_selector_to_method_id(cls, value: Any) -> Any:
        if isinstance(value, str) and not all(c in string.hexdigits for c in value):
            return keccak(text=value)[:4].hex()

        return value

    def __hash__(self) -> int:
        return int(self.accessory.lower().replace("0x", "") + self.method.hex(), 16)


class Accessory(ManagerAccessMixin):
    def __init__(self, address: AddressType | ContractInstance, *purses: "Purse"):
        self.address = self.conversion_manager.convert(address, AddressType)

        if isinstance(address, ContractInstance):
            self.contract = address

        # Installed purses, indexed by purse address
        self.purses: dict[AddressType, "Purse"] = {
            purse.address: purse for purse in purses
        }

    # TODO: `Accessory.load_package_type(package: uri or PackageManifest, contract_name: str)`

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.address}>"

    def __hash__(self) -> int:
        return int(self.address.lower().replace("0x", ""), 16)

    def __eq__(self, other: Any):
        if isinstance(other, Accessory):
            return self.address == other.address

        return self.conversion_manager.convert(other, AddressType) == self.address

    @cached_property
    def contract(self) -> ContractInstance:
        return self.chain_manager.contracts.instance_at(self.address)

    @cached_property
    def methods(self) -> list[AccessoryMethod]:
        """List of all methods required to install this accessory"""

        return [
            AccessoryMethod(method=abi.selector, accessory=self.address)
            for abi in self.contract.contract_type.abi
            if isinstance(abi, MethodABI)
        ]

    def install(self, bot):
        """
        Dynamically load and maintain the set of all Purse(s) using Accessory ``self``.

        Manages ``self.purses``, which is state of type ``set[Purse]``.
        """
        from .main import Purse
        from .package import MANIFEST

        PurseContractType = MANIFEST.Purse

        @bot.on_startup()
        async def load_purses_by_accessory(_ss):
            df = PurseContractType.AccessoryUpdated.query(
                "datetime,contract_address,old_accessory,new_accessory",
            )
            # TODO: Prune "replacements" (where a newer row undoes an older one)

            # NOTE: Use `set` for `O(1)` __contains__ check
            self.purses = {
                row.contract_address: Purse(row.contract_address) for row in df
            }

        @bot.on_(PurseContractType.AccessoryUpdated, old_accessory=self.address)
        async def remove_purse(log):
            if log.contract_address in self.purses:
                del self.purses[log.contract_address]

        @bot.on_(PurseContractType.AccessoryUpdated, new_accessory=self.address)
        async def add_purse(log):
            if log.contract_address not in self.purses:
                self.purses[log.contract_address] = Purse(log.contract_address)



from typing import TYPE_CHECKING, Any

from ape.contracts import (
    ContractInstance,
    ContractCallHandler,
    ContractTransactionHandler,
    ContractEvent,
)
from ape.utils import ManagerAccessMixin, cached_property, ZERO_ADDRESS
from ape.types import AddressType, ContractLog, HexBytes

if TYPE_CHECKING:
    from ape.api import AccountAPI
    from ape.api.address import BaseAddress
    from ape.api.transactions import ReceiptAPI

    from .accessory import Accessory


class Purse(ManagerAccessMixin):
    def __init__(
        self,
        account: "AccountAPI | BaseAddress | AddressType",
        *accessories: "Accessory",
    ):
        from ape.api import AccountAPI

        self.address = self.conversion_manager.convert(account, AddressType)

        if isinstance(account, AccountAPI):
            self.wallet = account

        self.accessories = set(accessories)

        # Installed accessories in wallet, indexed by method ID
        self._cached_accessories_by_method_id = {
            method: accy for accy in accessories for method in accy.methods
        }

    @classmethod
    def initialize(
        cls,
        account: "AccountAPI",
        *accessories: "Accessory",
        singleton: "AddressType | None" = None,
    ):
        assert singleton, "Needs support for package version"
        account.set_delegate(
            singleton,
            data=singleton.update_accessories.encode_input(
                [method.model_dump() for accy in accessories for method in accy.methods]
            ),
        )

        return cls(account, *accessories)

    @cached_property
    def wallet(self) -> "AccountAPI | None":
        if self.address in self.accounts_manager:
            return self.accounts_manager[self.address]

        return None

    @cached_property
    def contract(self) -> ContractInstance:
        # NOTE: Use `Purse` as proxy type
        return self.local_project.Purse.at(self.address)

    def _update_cache_from_logs(self, *logs: "ContractLog"):
        from purse.accessory import Accessory, AccessoryMethod

        for log in logs:
            if (
                log.contract_address == self.address
                and log.event_name == "AccessoryUpdated"
            ):
                old = AccessoryMethod(method=log.method, accessory=log.old_accessory)
                new = AccessoryMethod(method=log.method, accessory=log.new_accessory)

                if new.accessory != ZERO_ADDRESS:
                    try:
                        accessory = next(
                            accy
                            for accy in self.accessories
                            if accy.address == new.accessory
                        )

                    except StopIteration:
                        self.accessories.add(accessory := Accessory(new.accessory))

                    self._cached_accessories_by_method_id[new.method] = accessory

                elif old.accessory != ZERO_ADDRESS:
                    if old.method in self._cached_accessories_by_method_id:
                        del self._cached_accessories_by_method_id[old.method]

                    try:
                        accessory = next(
                            accy
                            for accy in self.accessories
                            if accy.address == old.accessory
                        )

                    except StopIteration:
                        continue

                    if all(
                        method.method not in self._cached_accessories_by_method_id
                        for method in accessory.methods
                    ):
                        self.accessories.remove(accessory)

    def add_accessories(
        self,
        *accessories: "Accessory",
        **txn_args,
    ) -> "ReceiptAPI":
        if not accessories:
            raise RuntimeError("Must provide at least one accessory")

        updates: list[dict] = [
            method.model_dump() for accy in accessories for method in accy.methods
        ]

        if "sender" not in txn_args and self.wallet:
            txn_args["sender"] = self.wallet

        receipt = self.contract.update_accessories(updates, **txn_args)

        self._update_cache_from_logs(*receipt.events)

        return receipt

    def remove_methods(
        self,
        *methods: "str | HexBytes",
        **txn_args,
    ) -> "ReceiptAPI":
        from purse.accessory import AccessoryMethod

        if not methods:
            raise RuntimeError("Must provide at least one accessory method")

        updates: list[dict] = [
            AccessoryMethod(accessory=ZERO_ADDRESS, method=method).model_dump()
            for method in methods
        ]

        if "sender" not in txn_args and self.wallet:
            txn_args["sender"] = self.wallet

        receipt = self.contract.update_accessories(updates, **txn_args)

        self._update_cache_from_logs(*receipt.events)

        return receipt

    def remove_accessories(
        self,
        *accessories: "Accessory",
        **txn_args,
    ) -> "ReceiptAPI":
        return self.remove_methods(
            *(m.method for accy in accessories for m in accy.methods),
            **txn_args,
        )

    def __getattr__(self, name: str) -> Any:
        if attr := getattr(self.contract, name, None):
            return attr

        for accy in self.accessories:
            if attr := getattr(accy.contract, name, None):
                match attr:
                    case ContractCallHandler():
                        return ContractCallHandler(
                            contract=self.contract, abis=attr.abis
                        )

                    case ContractTransactionHandler():
                        return ContractTransactionHandler(
                            contract=self.contract, abis=attr.abis
                        )

                    case ContractEvent():
                        return ContractEvent(contract=self.contract, abi=attr.abi)

        else:
            # NOTE: **Must raise** `AttributeError` if no attribute found, so that method bubbles up
            raise AttributeError(
                f"Method {attr} not a registered accessory method or event"
            )

    def install(self, bot):
        """
        Dynamically maintain the set of all accessories installed for ``self``.

        Manages internal cache of an instance ``self`` of this class.
        """
        from silverback.types import TaskType

        async def load_purses_by_accessory(snapshot):
            df = self.contract.AccessoryUpdated.query(
                "method,old_accessory,new_accessory"
            )
            self._update_cache_from_logs(*df)

        load_purses_by_accessory.__name__ = (
            f"purse:main:{load_purses_by_accessory.__name__}"
        )
        bot.broker_task_decorator(TaskType.STARTUP)(load_purses_by_accessory)

        async def update_accessory(log):
            self._update_cache_from_logs(log)

        update_accessory.__name__ = f"purse:main:{update_accessory.__name__}"
        bot.broker_task_decorator(
            TaskType.EVENT_LOG, container=self.contract_AccessoryUpdated
        )(update_accessory)



import pytest

from purse import Accessory, Purse


@pytest.fixture(scope="session")
def owner(accounts):
    return accounts[0]


@pytest.fixture(scope="session")
def other(accounts):
    return accounts[-1]


@pytest.fixture(scope="session")
def singleton(project, owner):
    return owner.deploy(project.Purse)


@pytest.fixture()
def purse(singleton, owner):
    # NOTE: Empty purse for testing purposes
    return Purse.initialize(owner, singleton=singleton)


@pytest.fixture(scope="session")
def create2_deployer(project, owner):
    return Accessory(owner.deploy(project.Create))


@pytest.fixture(scope="session")
def multicall(project, owner):
    return Accessory(owner.deploy(project.Multicall))


@pytest.fixture(scope="session")
def sponsor(project, owner):
    return Accessory(owner.deploy(project.Sponsor))


@pytest.fixture(scope="session")
def dummy(compilers, owner):
    SRC = """# pragma version 0.4.1
last_call: public(Bytes[2048])

@external
@payable
def __default__():
    self.last_call = slice(msg.data, 0, 2048)
    """
    container = compilers.compile_source("vyper", SRC, contractName="Dummy")
    return Accessory(container.deploy(sender=owner))



import pytest
from ape.utils import ZERO_ADDRESS

from purse import Purse


@pytest.fixture()
def purse(singleton, owner, create2_deployer):
    return Purse.initialize(owner, create2_deployer, singleton=singleton)


@pytest.fixture(scope="module")
def container(project):
    return project.Multicall  # NOTE: Just random choice


@pytest.fixture(scope="module")
def blueprint(container, owner):
    return owner.declare(container).contract_address


@pytest.mark.parametrize("salt", [b"", b"Custom Salt"])
def test_create_blueprint(purse, blueprint, salt):
    if len(salt) < 32:
        salt = salt + b"\x00" * (32 - len(salt))

    tx = purse.create(b"", blueprint, salt, sender=purse.wallet)

    assert tx.events == [
        purse.DeploymentFromBlueprint(blueprint=blueprint, salt=salt, args=b""),
    ]


@pytest.mark.parametrize("salt", [b"", b"Custom Salt"])
def test_raw_create(purse, container, salt):
    if len(salt) < 32:
        salt = salt + b"\x00" * (32 - len(salt))

    initcode = container.contract_type.get_deployment_bytecode()

    tx = purse.create(initcode, ZERO_ADDRESS, salt, sender=purse.wallet)

    assert tx.events == [
        purse.Deployment(salt=salt, initcode=initcode),
    ]



import click
from ape import project
from ape.cli import ConnectedProviderCommand, account_option, network_option


@click.group()
def cli():
    """Commands for deploying the system."""


@cli.command(cls=ConnectedProviderCommand)
@network_option()
@account_option()
@click.option("--publish", is_flag=True, default=False)
def purse(account, publish):
    account.deploy(project.Purse, publish=publish)


@cli.command(cls=ConnectedProviderCommand)
@network_option()
@account_option()
@click.option("--publish", is_flag=True, default=False)
def multicall(account, publish):
    account.deploy(project.Multicall, publish=publish)



from ape import convert, reverts
from ape.utils import ZERO_ADDRESS


def test_init(singleton, purse, owner):
    assert owner.delegate == singleton
    assert purse.address == owner.address
    assert len(purse.accessories) == 0


def test_can_transfer(purse, accounts):
    # NOTE: Make sure reentrancy doesn't lock us out of sending ether
    balance = purse.balance
    accounts[-1].transfer(purse.address, "1 ether")
    assert purse.balance - balance == convert("1 ether", int)


def test_add_rm_accessory(purse, dummy):
    assert dummy not in purse.accessories
    assert all(
        purse.contract.accessoryByMethodId(method.method) == ZERO_ADDRESS
        for method in dummy.methods
    )

    tx = purse.add_accessories(dummy, sender=purse.wallet)
    assert dummy in purse.accessories
    assert all(
        purse.contract.accessoryByMethodId(m.method) == dummy.address
        for m in dummy.methods
    )
    assert tx.events == [
        purse.contract.AccessoryUpdated(
            method=m.method,
            old_accessory=ZERO_ADDRESS,
            new_accessory=m.accessory,
        )
        for m in dummy.methods
    ]

    tx = purse.remove_accessories(dummy, sender=purse.wallet)
    assert dummy not in purse.accessories
    assert all(
        purse.contract.accessoryByMethodId(method.method) == ZERO_ADDRESS
        for method in dummy.methods
    )
    assert tx.events == [
        purse.contract.AccessoryUpdated(
            method=m.method,
            old_accessory=m.accessory,
            new_accessory=ZERO_ADDRESS,
        )
        for m in dummy.methods
    ]


def test_cant_call_arbitrary(purse):
    with reverts(message="Purse:!no-accessory-found"):
        purse.contract(data="0xa1b2c3d", sender=purse.wallet)



import pytest

from purse import Purse


@pytest.fixture()
def purse(singleton, owner, sponsor):
    return Purse.initialize(owner, sponsor, singleton=singleton)


def test_sponsor_plain_eth_transfer(purse, dummy, owner):
    # Generate random ETH amount > 0 (fuzz with hypothesis)
    # Call sponsor with no data and value > 0
    # Verify dummy received and processed ETH
    pass


def test_sponsor_eth_attached_to_call(purse, dummy, owner):
    # Fuzz random bytes data with hypothesis and random eth value
    # Sponsor a call with calldata and value
    # Check dummy recorded the data hash
    pass


def test_sponsor_target_is_purse_self(purse, owner):
    # Send a call to self with ETH, triggering __default__ (reentrant)
    # Expect successful processing (no revert, maybe check call depth if possible)
    pass


def test_sponsor_reverts_if_signature_expired(purse, owner):
    # Construct an expired signature using a timestamp in the past
    # Use ape reverts to check the Sponsor:!expired-signature error is bubbled up
    pass


def test_sponsor_reverts_if_nonce_used_is_incorrect(purse, owner):
    # Fetch the correct nonce
    # Loop and use wrong nonce (off-by-one)
    # Use ape reverts, check Sponsor:!unauthorized-signer erro
    pass


def test_sponsor_reverts_signer_is_unauthorized(purse, other):
    # Mallory signs instead of the owner
    # Use ape reverts, check Sponsor:!unauthorized-signer erro
    pass



import ape
import pytest

from purse import Purse


@pytest.fixture()
def purse(singleton, owner, multicall):
    return Purse.initialize(owner, multicall, singleton=singleton)


def test_empty_multicall(purse):
    purse.execute([], sender=purse.wallet)


def test_single_multicall(purse, accounts):
    a = accounts[1]
    bal_a = a.balance
    purse.execute(
        [dict(target=a, value="1 ether", data=b"")],
        sender=purse.wallet,
    )
    assert a.balance - bal_a == ape.convert("1 ether", int)


def test_many_multicall(purse, accounts):
    a, b, c = accounts[1:4]
    bal_a = a.balance
    bal_b = b.balance
    bal_c = c.balance

    purse.execute(
        [
            dict(target=a, value="1 ether", data=b""),
            dict(target=b, value="2 ether", data=b""),
            dict(target=c, value="3 ether", data=b""),
        ],
        sender=purse.wallet,
    )

    assert a.balance - bal_a == ape.convert("1 ether", int)
    assert b.balance - bal_b == ape.convert("2 ether", int)
    assert c.balance - bal_c == ape.convert("3 ether", int)


def test_only_owner_can_multicall(purse, other):
    assert purse.address != other.address

    with ape.reverts(message="Multicall:!authorize"):
        purse.execute(
            [dict(target=other, value="1 ether", data=b"")],
            sender=other,
        )
